#include "io.h"
#include "math.h"
#include "avrx.h"
#include "hardware.h"
#include "MotorTask.h"
#include "lcdio.h"
#include "adc_driver.h"
#include "maze.h"
#include "GP2D02.h"

#define D_POSITION  0       // Menu items, make monotonic.
#define D_ADC       1
#define D_FLOOR     2
#define D_MAZE      3
#define D_SHARP     4

#define FALSE 0
#define TRUE 1

Mutex MicrophoneFlag;
TimerControlBlock LcdTimer;
int MazeTimer;              // Timer for 3 minute contest timeout

extern Mutex lcdmutex;
    
static char hello[] FLASH   = "Dilbert II";
static char stX[] FLASH     = "X: ";
static char stR[] FLASH     = " R: ";
static char stY[] FLASH     = "Y: ";
static char stL[] FLASH     = " L: ";
static char stT[] FLASH     = "T: ";
static char sON[] FLASH     = "On ";
static char sOFF[] FLASH    = "Off";
static char sADC[] FLASH    = "ADC";
static char sFloor[] FLASH  = "LineTrack";
static char sFloor2[] FLASH = "Value: ";
static char sMaze[] FLASH   = "Maze Options";
static char sReset[] FLASH  = "Reset";
static char sSharp[] FLASH  = "GP2DO2";
static char sHalt[] FLASH   = "Fatal Error ";


// Helper for ADC display

void PutChannel(int p)
{
    lcd_PutHexWord(ADC_Channel(p));
    lcd_put_char(' ');
}

void HaltError(char ploc)
{
    
    lcd_PrintMessageAt(sHalt, LCD_LINE1);
    lcd_PutHexByte(ploc);
    AvrXHalt();
}


void ResetDisplay(void)
{
    lcd_wr_inst(LCD_CLEAR);
    lcd_wr_inst(LCD_HOME);
    lcd_wr_inst(LCD_DISPLAY | LCD_D | LCD_C);
//    AvrXSetSemaphore(&lcdmutex);
    lcd_putstr(hello);
    lcd_setcursor(LCD_LINE1 + 11);
}
/*
    LcdTask updates the LCD display 5 times/second.

    Converts internal numbers into .1" and degrees

*/
AVRX_TASKDEF(LcdTask, 10, 3)
{
    int8_t  mic_on = 0,
            mic_off = 0,
            led = 0,
            mode = D_POSITION,
            option = 0,
            mode_sw = FALSE;
    int16_t tmp;

    ADC_Init();
    
    lcd_4_init();
    lcd_wr_inst(LCD_FUNC | LCD_N);
    ResetDisplay();
    Init_GP2D02();

    while(1)
    {
        AvrXDelay(&LcdTimer, MILLISECOND(100));

        if (++led == 8)                     // LED on for 200ms/1 second
        {
            sbi(PORTB, LED);
        }
        else if (led == 10)
        {
            led = 0;
            cbi(PORTB, LED);
        }

        // Update display on odd ticks (200ms)
        
        if (led & 0x01)
        {
            switch(mode)
            {
                case D_POSITION:
                    lcd_setcursor(LCD_LINE2);
                    lcd_putstr(stX);
                    lcd_PutHexWord(iX);         // Scale for 10th of an inch
                    lcd_putstr(stR);
                    lcd_PutHexWord(Right.Encoder);
                    lcd_EraseToEndOfLine();
                    lcd_setcursor(LCD_LINE3);
                    lcd_putstr(stY);
                    lcd_PutHexWord(iY);
                    lcd_putstr(stL);
                    lcd_PutHexWord(Left.Encoder);
                    lcd_EraseToEndOfLine();
                    lcd_setcursor(LCD_LINE4);
                    lcd_putstr(stT);
                    lcd_PutHexWord(iTheta);
                    lcd_EraseToEndOfLine();
                    break;
                case D_ADC:
                    lcd_setcursor(LCD_LINE2);
                    lcd_putstr(sADC);
                    lcd_EraseToEndOfLine();
                    lcd_setcursor(LCD_LINE3);
                    PutChannel(0);
                    PutChannel(1);
                    PutChannel(2);
                    PutChannel(3);
                    lcd_setcursor(LCD_LINE4);
                    PutChannel(4);
                    PutChannel(5);
                    PutChannel(6);
                    PutChannel(7);
                    break;
                case D_FLOOR:
                    tmp = GetLineTrackInfo(12);
                    lcd_setcursor(LCD_LINE2);
                    lcd_putstr(sFloor);
                    lcd_EraseToEndOfLine();
                    lcd_setcursor(LCD_LINE3);
                    lcd_EraseToEndOfLine();
                    lcd_setcursor(LCD_LINE3+tmp+9);
                    lcd_put_char('*');
                    lcd_setcursor(LCD_LINE4);
                    lcd_putstr(sFloor2);
                    lcd_PutHexByte(tmp);
                    lcd_EraseToEndOfLine();
                    break;
                case D_MAZE:
                    ResetDisplay();
                    lcd_setcursor(LCD_LINE2);
                    lcd_putstr(sMaze);
                    if (option)
                    {
                        option = 0;
                        WritepNode(0);
                        lcd_putstr(sReset);
                    }
                    break;
                case D_SHARP:
                    Read_GP2D02();
                    lcd_setcursor(LCD_LINE2);
                    lcd_putstr(sSharp);
                    lcd_EraseToEndOfLine();
                    lcd_setcursor(LCD_LINE3);
                    lcd_PutHexByte(LeftRaw);
                    lcd_EraseToEndOfLine();
                    lcd_setcursor(LCD_LINE3+4);
                    lcd_PutHexByte(FrontRaw);
                    lcd_setcursor(LCD_LINE3+8);
                    lcd_PutHexByte(RightRaw);
                    lcd_setcursor(LCD_LINE4);
                    lcd_EraseToEndOfLine();
                    break;
                default:
                    mode = D_POSITION;      // Reset mode
                    break;
            }
        }
        
        // Evaluate Switch Input

        if (inp(PINB) & BV(SW))    // Switch is low active
        {
            if (TAP < mode_sw && mode_sw < LONG)
                ++mode;
            else if (LONG < mode_sw && mode_sw < RESET)
                ++option;
            else if (mode_sw >= RESET)
                asm( \
                    "cli\n" \
                    "\tclr R30\n" \
                    "\tclr R31\n" \
                    "\tijmp\n" \
                    );

            mode_sw = 0;
        }
        else
        {
            ++mode_sw;
            option = 0;
        }

        // Evaluate Audio Detect Signal

        if (inp(PINC) & BV(PC0))
        {
            if (++mic_on >= MIC_THRESHOLD)
            {
                if (!mic_off)
                {
                    MazeTimer = 0;      // Zero out to halt current operations
                    AvrXSetSemaphore(&MicrophoneFlag);  // Trigger any pending tasks
                }
                lcd_setcursor(LCD_LINE1 + 17);
                lcd_putstr(sON);
                mic_on = MIC_THRESHOLD;
            }
            mic_off = 0;
        }
        else
        {
            if (++mic_off >= MIC_THRESHOLD)
            {
                // Clear any pending flag
                if (!mic_on)
                    (void)AvrXTestSemaphore(&MicrophoneFlag);
                lcd_setcursor(LCD_LINE1 + 17);
                lcd_putstr(sOFF);
                mic_off = MIC_THRESHOLD;
            }
            mic_on = 0;
        }

        // Hokey timer for maze running (1/10's of a second)

        if (MazeTimer)
            MazeTimer--;
    }
}

