#include <math.h>
#include <eeprom.h>
#include "avrx.h"
#include "hardware.h"
#include "MotorTask.h"

/*
    Globals/Statics etc.
*/

MotorInfo Left, Right;
float   X, Y, Theta;
int16_t iX, iY, iTheta;
static TimerControlBlock MotorTimer;
Mutex MotorLoop;

// Initialized data stored in EEPROM (tuning parameters)

char  Kp EEPROM = 16;       // Proportional gain
char  Kd EEPROM = 12;       // Derivative gain
char  Td EEPROM = 1;        // Derivative period
char  Ki EEPROM = 0;        // Integral gain
char  Ko EEPROM = 1;        // Output factor
int   Acceleration EEPROM = 0x70;

void MotorInit(void)
{
    // Initialize Motion Control data
    
    Left.Kp = Right.Kp = eeprom_rb((unsigned)&Kp);
    Left.Kd = Right.Kd = eeprom_rb((unsigned)&Kd);
    Left.Ki = Right.Ki = eeprom_rb((unsigned)&Ki);
    Left.Ko = Right.Ko = eeprom_rb((unsigned)&Ko);
    Left.Td = Right.Td = eeprom_rb((unsigned)&Td);
    Left.Acceleration = Right.Acceleration = eeprom_rw((unsigned)&Acceleration);

}

uint16_t DoPid(MotorInfo *p)
{
    int16_t Perror, output;

    Perror = (p->Setpoint>>8) - p->Encoder;     // Set point is an 24.8 number (+/- 1 mile..)

    // Derivative error is the delta Perror over Td

    output = (p->Kp*Perror + p->Kd*(Perror - p->PrevErr) + p->Ki*p->Ierror)/p->Ko;
    p->PrevErr = Perror;

    // Accumulate Integral error *or* Limit output.
    // Stop accumulating when output saturates
    
    if (output >= MAXOUTPUT)
        output = MAXOUTPUT;
    else if (output <= -MAXOUTPUT)
        output = -MAXOUTPUT;
    else
        p->Ierror += Perror;

    // Convert output to 0xFF-0x7F-0x00 (Fwd, netural, Rev)
    // locked anti-phase drive.
        
    return (output - (MAXOUTPUT+1));
}

/*+ ---------------------------------------------------------
 void DoMotion(MotorInfo *p)

 Called at the loop rate to add "velocity" to the setpoint thus
 effecting a motion.

 Velocity is ramped up and down by "acceleration"

 Velocity and Acceleration are 8.8 numbers.  Velocity
 is added to the setpoint (an 24.8 number)
-*/

void DoMotion(MotorInfo *p)
{
    if (p->Velocity < p->VelocitySetpoint)
    {
        p->Velocity += p->Acceleration;
        if (p->Velocity > p->VelocitySetpoint)
            p->Velocity = p->VelocitySetpoint;
    }
    else 
    {
        p->Velocity -= p->Acceleration;
        if (p->Velocity < p->VelocitySetpoint)
            p->Velocity = p->VelocitySetpoint;
    }
    AddToPosition(p, p->Velocity);
}


/*+ -------------------------------------------------------------
    void AddToPosition(pMotorInfo, int)

    Add an 8.8 value to the current setpoint.

-*/

void AddToPosition(pMotorInfo p, int16_t o)
{
    char s = inp(SREG);
    BeginCritical();
    p->Setpoint += o;        // Set point is 24.8 number
    outp(s, SREG);
}

AVRX_TASKDEF(MotorTask, 28, 2)
{
    int8_t LeftEnc, RightEnc;
    float LocalTheta, DeltaTheta, dist;
    extern int8_t RightEncoder, LeftEncoder;

    MotorInit();

    while(1)
    {
        AvrXStartTimer(&MotorTimer, MILLISECOND(10));
        AvrXSetSemaphore(&MotorLoop);   // Synchronize other motion tasks
        
        BeginCritical();
        LeftEnc = LeftEncoder;
        LeftEncoder = 0;
        RightEnc = RightEncoder;
        RightEncoder = 0;
        EndCritical();

        Left.Encoder += LeftEnc;
        Right.Encoder += RightEnc;

        // Do the odometery calculations
       
        DeltaTheta = (LeftEnc-RightEnc) * (DEG_PER_TICK/2.0);
        dist = (LeftEnc+RightEnc) * (DIST_PER_TICK/2.0);

        LocalTheta = Theta + DeltaTheta;
        X -= dist * cos(LocalTheta);
        Y -= dist * sin(LocalTheta);
        LocalTheta += DeltaTheta;
        
        if (LocalTheta > 2.0*M_PI)
            LocalTheta -= 2.0*M_PI;       // Keep theta between 0-360
        else if (LocalTheta < 0.0)
            LocalTheta += 2.0*M_PI;

        // Keep integer versions around for user interface & Maze running
        
        iX = X * 10.0;
        iY = Y * 10.0;
        iTheta = Theta * (360 / (2*M_PI));

        DoMotion(&Left);
        outp(DoPid(&Left), LEFT_DRIVEL);
        DoMotion(&Right);
        outp(DoPid(&Right), RIGHT_DRIVEL);

        AvrXWaitTimer(&MotorTimer);
    }
}
