#include <io.h>
#include <avrx.h>
#include <interrupt.h>
#include <math.h>
#include <eeprom.h>
#include "maze.h"
#include "adc_driver.h"
#include "hardware.h"
#include "MotorTask.h"
#include "LcdTask.h"

#define TRUE    1
#define FALSE   0

Node EEPROM eeNodeData[MAXNODES];   // Solved Maze

NodeFlags   branch,         // New Node Info from FollowLine
            CurrentNode;    // Current Node we are exploring
Node        *pNode;         // Pointer into EEPROM stack
uint8_t     BackTrack,      // Flag set while pruning
            solving;        // Flag


/*+ -----------------------------------------------------------------
    int FollowLine(void)
    
    Follows line, returning status
        0 = Lost line
        1, 2 or 3 = intersection of N branches
        4 = Goal (end of maze)

    Fills in "Results" with bits showing which branch detected.

-*/
int FollowLine(void)
{
    int d, retc;

    // Start Motion (see MotorCtl.c), track line until either line dissapears or side
    // branches detected.
    
    Left.VelocitySetpoint = Right.VelocitySetpoint = FloorData.iFollowVelocity;

    while(1)
    {
        if ((retc = SynchWithMotion()))
            return retc;
        
        DoFloorThreshold();
        
        if (Sensor.b.left_present || Sensor.b.right_present || !Sensor.b.center_present)
            break;

        // Adjust track based upon actual velocity, not setpoint velocity, so that
        // we can get aligned while accelerating...
            
        d = GetLineTrackInfo(Left.Velocity/FloorData.iFloorScale);
        AddToPosition(&Left,   d);  // Bias drive to correct error
        AddToPosition(&Right, -d);
    }

    // Advance 1/2", or until side sensors detect a branch.

    d = FollowLineHalfInch;

    while(d--)
    {
        DoFloorThreshold();
        if (Sensor.b.left_present || Sensor.b.right_present)
            break;

        if ((retc = SynchWithMotion()))
            return retc;
    }
        
    // Now, advance 3.5 inches (from side sensor to wheel centerline)
    // taking in account a roughly 1.4" stopping distance

    d = FollowLineThreeHalfInch;

    branch.byte = 0;                    // Reset all bits except for center line.
    branch.b.center_present=TRUE;
    
    while(d--)
    {
        DoFloorThreshold();

        if (Sensor.b.left_present)
            branch.b.left_present = TRUE;
        if (Sensor.b.right_present)
            branch.b.right_present = TRUE;
        if (!Sensor.b.center_present)
            branch.b.center_present = FALSE;

        if ((retc = SynchWithMotion()))
            return retc;
    }

    Left.VelocitySetpoint = 0;      // Ok, stop motion (actually, start deceleration
    Right.VelocitySetpoint = 0;

    
    retc = DEAD_END;                // 0x00
    
    if (branch.b.left_present)
        retc++;
    if (branch.b.right_present)
        retc++;
    if (branch.b.center_present)
        retc++;

    // if either side branches see something, then they must be over the goal.
        
    if (Sensor.b.left_present | Sensor.b.right_present)
        retc = GOAL;
        
    return (retc);
}

void WaitStart(void)
{
    AvrXWaitSemaphore(&MicrophoneFlag);
}

char PruneTree( void)
{
    char NeedToTurnLeft = FALSE;
    
    BackTrack = FALSE;
    switch(CurrentNode.b.branch)   // Prune dead branch
    {
        case LEFT_BRANCH:
            CurrentNode.b.left_visited = CurrentNode.b.left_present = 0;
            break;
        case RIGHT_BRANCH:
            CurrentNode.b.right_visited = CurrentNode.b.right_present = 0;
            break;
        case CENTER_BRANCH:
            CurrentNode.b.center_visited = CurrentNode.b.center_present = 0;
            break;
        default:
            HaltError(INVALID_BRANCH0);
    }
    
    // Select next branch (either center or right), or delete node if
    // no branches left to explore.

    if (CurrentNode.b.center_present && !CurrentNode.b.center_visited)
    {
        NeedToTurnLeft = TRUE;
        CurrentNode.b.center_visited = TRUE;   // branch must be LEFT
        CurrentNode.b.branch = CENTER_BRANCH;
    }
    else if (CurrentNode.b.right_present && !CurrentNode.b.right_visited)
    {
        if (CurrentNode.b.branch == CENTER_BRANCH)    // Could either be LEFT or CENTER branch
            NeedToTurnLeft = TRUE;
        CurrentNode.b.branch = RIGHT_BRANCH;
        CurrentNode.b.right_visited = TRUE;
    }
    else    // Delete node, can be only center or right branch we came from...
    {
        BackTrack = TRUE;
        WritepNode(--pNode);

        if (CurrentNode.b.branch == RIGHT_BRANCH)
            NeedToTurnLeft = TRUE;
            
        if (pNode < &eeNodeData[0])
            solving = FALSE;
        else
            CurrentNode.byte = eeprom_rb((unsigned)pNode);
    }
    WriteNode(pNode, CurrentNode);

    return NeedToTurnLeft;
}

/*
 When solving maze, use left hand rule.  At each intersection
 take the left most passage.  When moving forward, add nodes
 when backtracking, delete nodes that have all arms visited.
 */
AVRX_TASKDEF(SolveMaze, 30, 4)
{
    Node *LastNode;

    InitFloorParameters();

    // Initial state is eepNode = 0 and pNode = &eeNodeData.  After first node encountered
    // eepNode = &eeNodeData.  When running maze just follow last branch taken for each node.
    // When pNode > eepNode, then write new node for each node encountered.  When backtracking
    // write out new pNode for each node deleted.
    
    while(1)    // Run through nodes already stored (if any)
    {
        WaitStart();

        pNode = (Node *)&eeNodeData;    // Start at beginning of nodelist
        LastNode = (Node*)eeprom_rw((unsigned)&eepNode);
        
        BackTrack = FALSE;
        solving = TRUE;
        CurrentNode.byte = 0;

        MazeTimer = MAZE_TIMEOUT;   // Decremented in LCD Task

        // While pNode < Last solved pNode, then just run nodes
        // After reading in last node in list, exit and start solving

        while(solving && (pNode < LastNode))
        {
            switch(FollowLine())
            {
                case 1:
                    if (branch.b.left_present)
                        Turn(LEFT, TO_CENTER_LINE);
                    else
                        Turn(RIGHT, TO_CENTER_LINE);
                    break;
                case 2:         // Read from EEPROM which branch to take.
                case 3:
                    CurrentNode.byte = eeprom_rb((unsigned)++pNode); // [BUG? Post Increment]
                    switch(CurrentNode.b.branch)
                    {
                        case LEFT_BRANCH:
                            Turn(LEFT, TO_CENTER_LINE);
                            break;
                        case CENTER_BRANCH:
                            break;
                        case RIGHT_BRANCH:
                            Turn(RIGHT, TO_CENTER_LINE);
                            break;
                        default:
                            HaltError(INVALID_BRANCH1);
                            break;
                    }
                    break;
                case GOAL:
                    solving = FALSE;
                    break;
                default:    // case TIMEOUT or DEAD_END
                    Left.VelocitySetpoint = 0;   // Ok, stop motion (actually, start deceleration
                    Right.VelocitySetpoint = 0;
                    HaltError(TIMEOUT_DEADEND);      // Maze unsolvable in time allowed or logic error.
                break;
            }
        }

        // Now, continue running maze until goal or 3 minutes up.
        
        while(solving)
        {
            switch(FollowLine())    // Returns # branches, particulars in "branch"
            {
                case DEAD_END:      // Dead End, turn around and backtrack
                    Turn(BACK, TO_CENTER_LINE);
                    BackTrack = TRUE;
                    break;
                case 1:             // Not an intesection, just turn and continue
                    if (branch.b.left_present)
                        Turn(LEFT, TO_CENTER_LINE);
                    else
                        Turn(RIGHT, TO_CENTER_LINE);
                    break;
                case 2:
                case 3:             // 2 or more branches is an intersection
                    if (BackTrack == TRUE)
                    {
                        if (PruneTree())
                            Turn(LEFT, TO_CENTER_LINE);
                    }
                    else    // New Node!, write out old one.
                    {
                        WriteNode(pNode, CurrentNode);
                        if (++pNode > &eeNodeData[MAXNODES])
                        {
                            HaltError(STACK_FULL);
                        }

                        CurrentNode.byte = branch.byte;
            
                        // by definition two or more branches, so
                        // there must be either a left or a center branch

                        if (CurrentNode.b.left_present)
                        {
                            Turn(LEFT, TO_CENTER_LINE);
                            CurrentNode.b.left_visited = TRUE;
                            CurrentNode.b.branch = LEFT_BRANCH;
                        }
                        else if (CurrentNode.b.center_present)
                        {
                            CurrentNode.b.center_visited = TRUE;
                            CurrentNode.b.branch = CENTER_BRANCH;
                        }
                        else
                        {
                            HaltError(INVALID_BRANCH2);
                        }
                    }
                    break;
                case GOAL:
                    (void) WriteNode(pNode, CurrentNode);
                    solving = FALSE;
                    break;
                case TIMEOUT:
                    Left.VelocitySetpoint = 0;      // Ok, stop motion (actually, start deceleration
                    Right.VelocitySetpoint = 0;
                    
                    while ((BackTrack == TRUE) && (solving == TRUE))
                    {
                        PruneTree();  // Walk tree pruning dead branches.
                    }
                    solving = FALSE;
                    break;
                default:
                    HaltError(INVALID_FOLLOWLINE);
            }
        }
    }               // End while(1)...
}                   // end of task...


