#include <io.h>
#include <avrx.h>
#include <math.h>
#include <eeprom.h>
#include "maze.h"
#include "adc_driver.h"
#include "hardware.h"
#include "MotorTask.h"
#include "serialio.h"

#define TRUE 1
#define FALSE 0

// EEPROM tuning paramters

FloorSensorParameters EEPROM eeFloorData =
{
    FloorScale,             // 8.8
    FollowVelocity,         // 8.8 Number
    FloorThresholdLow,
    FloorThresholdHigh
};

int EEPROM eepNode;                 // Index of current node

FloorSensorParameters FloorData;
NodeFlags   Sensor;     // Holding register for sensor logic level

/*+ -------------------------------------------------------------
    void SynchWithMotion(void)

    Synchronize the calling task with the motor control loop
-*/

int SynchWithMotion(void)
{
    AvrXWaitSemaphore(&MotorLoop);
    if (MazeTimer == 0)
        return TIMEOUT;
    else
        return FALSE;
}

/*+ -----------------------------------------------------------------

    InitFloorParameters()
    
-*/
void InitFloorParameters(void)
{
    eeprom_read_block((void*)&FloorData, (unsigned)&eeFloorData, sizeof(FloorData));
}

/*+ -----------------------------------------------------------------
    int GetLineTrackInfo(int scale)

    Passed: Integer scale factor XX.XX
    Returns: 8.8 number from 0.00 to +/- Scale
    Note:   10 bit ADC value, hence divisor of 0x3FF (>>10)
-*/
int GetLineTrackInfo(int scale)
{
    long foo = (ADC_Channel(LEFTLINE) - ADC_Channel(RIGHTLINE));
    foo *= scale;
    
    return (foo>>10);
}

/*+ -----------------------------------------------------------------

    void Turn(int)

    Turn robot int degrees.  Negative numbers turn right (cw)
    posative numbers turn left (ccw)

    If bCenter is set, then Turn looks for the center line to
    dissapear and then re-appear.  It will terminate the turn when
    the re-appeared line is centered under the sensors.
-*/

void Turn(int iTurn, char bCenter)
{
    char CenterPresent = 0;
#ifdef _MONITOR_
    putchar('\r');
    putchar('T');
    PutHexWord(iTurn);
    putchar(' ');
#endif
    WaitForMotionToStop();
    
    if (iTurn > 0)                          // Start new motion
    {
        Left.VelocitySetpoint = -ONE_DEG/2;
        Right.VelocitySetpoint = ONE_DEG/2;
    }
    else
    {
        Left.VelocitySetpoint =   ONE_DEG/2;
        Right.VelocitySetpoint = -ONE_DEG/2;
        iTurn = -iTurn;
    }

    while (iTurn--)
    {
        SynchWithMotion();
        
        if (bCenter == TRUE)
        {
            DoFloorThreshold();

            switch(CenterPresent)
            {
                case 45:
                    if (Sensor.b.center_present == FALSE)
                        CenterPresent++;
                    break;
                case 46:
                    if (Sensor.b.center_present == TRUE)
                        iTurn = 0;
                    break;
                default:
                    CenterPresent++;        // Delay checking for 40 deg.
                    break;
            }
        }
#ifdef _MONITOR_
        PutHexByte(Sensor.byte);
        putchar(' ');
#endif
    }
    Left.VelocitySetpoint = 0;              // Stop motion
    Right.VelocitySetpoint = 0;

    WaitForMotionToStop();
}

/*+ -----------------------------------------------------------------

    void WaitForMotionToStop(void)

    Self Evident...

-*/

void WaitForMotionToStop(void)
{
    while (Left.Velocity)
        SynchWithMotion();
}

/*+ -----------------------------------------------------------------
    void DoThresholds(void)

    Process sensor inputs and performs hysteresis setting bits

-*/
void DoFloorThreshold(void)
{
    int t;

    t = ADC_Channel(LEFTBRANCH);
    if (Sensor.b.left_present && t < FloorData.iThresholdLow)
        Sensor.b.left_present = FALSE;
    else if (t > FloorData.iThresholdHigh)
        Sensor.b.left_present = TRUE;
        
    t = ADC_Channel(RIGHTBRANCH);
    if (Sensor.b.right_present && t < FloorData.iThresholdLow)
        Sensor.b.right_present = FALSE;
    else if (t > FloorData.iThresholdHigh)
        Sensor.b.right_present = TRUE;

    // Hysteresis too hard for center group.  Just use simple logic:
    // if any are high, then just call the group high.

    if (ADC_Channel(LEFTLINE) > FloorData.iThresholdLow ||
        ADC_Channel(CENTERLINE) > FloorData.iThresholdLow ||
        ADC_Channel(RIGHTLINE) > FloorData.iThresholdLow )
        Sensor.b.center_present = TRUE;
    else
        Sensor.b.center_present = FALSE;
}


// Write node data to the EEPROM.

void WriteNode(Node *pNode, NodeFlags CurrentNode)
{
    eeprom_wb((unsigned)&pNode->flags.byte, CurrentNode.byte);
    eeprom_wb((unsigned)&pNode->iX, iX);
    eeprom_wb((unsigned)&pNode->iY, iY);
    eeprom_wb((unsigned)&pNode->iTheta, iTheta);
    WritepNode(pNode);
}

// This does a word write...  There will be a 4ms delay for
// the previous write to complete

void WritepNode(Node *uNode)
{
    eeprom_wb((unsigned)&eepNode, (char)(unsigned)uNode);
    eeprom_wb(((unsigned)&eepNode)+1, (char)((unsigned)uNode>>8));
}
